"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typeSafeCssModulesPlugin = void 0;
const source_map_1 = require("source-map");
const Stylesheet_js_1 = require("./Stylesheet.js");
const node_path_1 = __importDefault(require("node:path"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const postcss_modules_1 = __importDefault(require("postcss-modules"));
const typeSafeCssModulesPlugin = (opts = {}) => {
    const cssModules = (0, postcss_modules_1.default)({
        generateScopedName: opts.generateScopedName,
        localsConvention: (className) => className,
        getJSON: async (from, exportTokens, to) => {
            if (to == null) {
                throw new Error("Could not determine CSS module output file location");
            }
            // Index the stylesheet so that we can lookup lines & columns
            const stylesheet = new Stylesheet_js_1.Stylesheet(await fs_extra_1.default.readFile(from, "utf8"));
            const cssModuleExports = [...Object.entries(exportTokens)].map(([tokenName, tokenValue]) => {
                // postcss-modules automatically picks the last declared name to use in exportTokens
                // In a scenario where an exported variable and a class name shares the same name,
                // this can be unexpected so we manually check for conflicts and throw in this scenario
                const classNameLocation = stylesheet.findClass(tokenName);
                const variableLocation = stylesheet.findVariable(tokenName);
                if (classNameLocation != null && variableLocation != null) {
                    throw new Error(`The class name ".${tokenName}" conflicts with an exported variable name with the same name\n    at ${from}:${variableLocation.line}:${variableLocation.column}`);
                }
                return {
                    localVariableName: tokenName,
                    value: tokenValue,
                    variableName: toCamelCase(tokenName),
                    location: classNameLocation ?? variableLocation,
                };
            });
            const sourceFile = node_path_1.default.relative(node_path_1.default.dirname(to), from);
            // Generate the JavaScript file
            const js = new source_map_1.SourceNode(null, null, sourceFile, [
                `import "./${node_path_1.default.basename(to)}";\n\n`,
                ...cssModuleExports.flatMap(el => [
                    new source_map_1.SourceNode(null, null, sourceFile, `export const `),
                    new source_map_1.SourceNode(el.location?.line ?? null, el.location?.column ?? null, sourceFile, `${el.variableName} = "${el.value}";\n`),
                ]),
                `export default { ${cssModuleExports
                    .map(el => el.variableName)
                    .join(", ")} };\n`,
            ]).toStringWithSourceMap();
            // Generate the TypeScript .d.ts file
            const dts = new source_map_1.SourceNode(null, null, sourceFile, [
                ...cssModuleExports.flatMap(el => {
                    return [
                        new source_map_1.SourceNode(null, null, sourceFile, `export declare const `),
                        new source_map_1.SourceNode(el.location?.line ?? null, el.location?.column ?? null, sourceFile, `${el.variableName}: string;\n`),
                    ];
                }),
                "\n",
                `declare const `,
                new source_map_1.SourceNode(1, 0, sourceFile, `__defaultExports: {\n`),
                ...cssModuleExports.flatMap(el => [
                    ` `,
                    new source_map_1.SourceNode(el.location?.line ?? null, el.location?.column ?? null, sourceFile, `${el.variableName}: string;\n`),
                ]),
                `};\n`,
                `export default __defaultExports;`,
            ]).toStringWithSourceMap({ file: `${node_path_1.default.basename(from)}.d.ts` });
            // Write everything out to disk
            await Promise.all([
                fs_extra_1.default.outputFile(node_path_1.default.join(node_path_1.default.dirname(to), `${node_path_1.default.basename(from)}.js`), js.code),
                fs_extra_1.default.outputFile(node_path_1.default.join(node_path_1.default.dirname(to), `${node_path_1.default.basename(from)}.js.map`), js.map.toString()),
                fs_extra_1.default.outputFile(node_path_1.default.join(node_path_1.default.dirname(to), `${node_path_1.default.basename(from)}.d.ts`), dts.code),
                fs_extra_1.default.outputFile(node_path_1.default.join(node_path_1.default.dirname(to), `${node_path_1.default.basename(from)}.d.ts.map`), dts.map.toString()),
            ]);
        },
    });
    const previousOnceExitFunction = cssModules.OnceExit;
    cssModules.OnceExit = (root, helper) => {
        // First do whatever postcss-modules may have configured to prevent dropping behavior
        previousOnceExitFunction?.(root, helper);
        // Now rename the file as a last action
        const originalOutputFile = helper.result.opts.to;
        if (originalOutputFile?.endsWith(".module.css")) {
            // We need to output `.css` since the generated .js file will import it
            // This is strictly more correct, since after compiling the sass module it is no longer a CSS module
            helper.result.opts.to = replaceModuleExtension(originalOutputFile);
        }
    };
    return cssModules;
};
exports.typeSafeCssModulesPlugin = typeSafeCssModulesPlugin;
exports.typeSafeCssModulesPlugin.postcss = true;
function toCamelCase(str) {
    return str.replace(/-./g, x => x[1].toUpperCase());
}
function replaceModuleExtension(filePath) {
    const dirname = node_path_1.default.dirname(filePath);
    const originalFileName = node_path_1.default.basename(filePath);
    // Output with a leading underscore in the hopes of preventing conflicts with non-module files
    // The hope is that Sass treats .scss files with leading underscores as partials and won't compile them to CSS,
    // so the chances of conflicts should be low.
    return `${dirname}/_${originalFileName.replace(".module.css", ".css")}`;
}
